@@ -13,9 +13,7 @@
 #include "utils.h"
 
 #ifdef HAVE_PTHREAD
-#define THREAD_BLOCK_SIZE 1024
 #include <pthread.h>
-static pthread_mutex_t g_seq_lock = PTHREAD_MUTEX_INITIALIZER;
 #endif
 
 gap_opt_t *gap_init_opt()
@@ -51,22 +49,22 @@ int bwa_cal_maxdiff(int l, double err, double thres)
 }
 
 // width must be filled as zero
-static int bwt_cal_width(const bwt_t *rbwt, int len, const ubyte_t *str, bwt_width_t *width)
+static int bwt_cal_width(const bwt_t *bwt, int len, const ubyte_t *str, bwt_width_t *width)
 {
 	bwtint_t k, l, ok, ol;
 	int i, bid;
 	bid = 0;
-	k = 0; l = rbwt->seq_len;
+	k = 0; l = bwt->seq_len;
 	for (i = 0; i < len; ++i) {
 		ubyte_t c = str[i];
 		if (c < 4) {
-			bwt_2occ(rbwt, k - 1, l, c, &ok, &ol);
-			k = rbwt->L2[c] + ok + 1;
-			l = rbwt->L2[c] + ol;
+			bwt_2occ(bwt, k - 1, l, c, &ok, &ol);
+			k = bwt->L2[c] + ok + 1;
+			l = bwt->L2[c] + ol;
 		}
 		if (k > l || c > 3) { // then restart
 			k = 0;
-			l = rbwt->seq_len;
+			l = bwt->seq_len;
 			++bid;
 		}
 		width[i].w = l - k + 1;
@@ -77,12 +75,11 @@ static int bwt_cal_width(const bwt_t *rbwt, int len, const ubyte_t *str, bwt_wid
 	return bid;
 }
 
-void bwa_cal_sa_reg_gap(int tid, bwt_t *const bwt[2], int n_seqs, bwa_seq_t *seqs, const gap_opt_t *opt)
+void bwa_cal_sa_reg_gap(int tid, bwt_t *const bwt, int n_seqs, bwa_seq_t *seqs, const gap_opt_t *opt)
 {
-	int i, max_l = 0, max_len;
+	int i, j, max_l = 0, max_len;
 	gap_stack_t *stack;
-	bwt_width_t *w[2], *seed_w[2];
-	const ubyte_t *seq[2];
+	bwt_width_t *w, *seed_w;
 	gap_opt_t local_opt = *opt;
 
 	// initiate priority stack
@@ -92,57 +89,40 @@ void bwa_cal_sa_reg_gap(int tid, bwt_t *const bwt[2], int n_seqs, bwa_seq_t *seq
 	if (local_opt.max_diff < local_opt.max_gapo) local_opt.max_gapo = local_opt.max_diff;
 	stack = gap_init_stack(local_opt.max_diff, local_opt.max_gapo, local_opt.max_gape, &local_opt);
 
-	seed_w[0] = (bwt_width_t*)calloc(opt->seed_len+1, sizeof(bwt_width_t));
-	seed_w[1] = (bwt_width_t*)calloc(opt->seed_len+1, sizeof(bwt_width_t));
-	w[0] = w[1] = 0;
+	seed_w = (bwt_width_t*)calloc(opt->seed_len+1, sizeof(bwt_width_t));
+	w = 0;
 	for (i = 0; i != n_seqs; ++i) {
 		bwa_seq_t *p = seqs + i;
 #ifdef HAVE_PTHREAD
-		if (opt->n_threads > 1) {
-			pthread_mutex_lock(&g_seq_lock);
-			if (p->tid < 0) { // unassigned
-				int j;
-				for (j = i; j < n_seqs && j < i + THREAD_BLOCK_SIZE; ++j)
-					seqs[j].tid = tid;
-			} else if (p->tid != tid) {
-				pthread_mutex_unlock(&g_seq_lock);
-				continue;
-			}
-			pthread_mutex_unlock(&g_seq_lock);
-		}
+		if (i % opt->n_threads != tid) continue;
 #endif
 		p->sa = 0; p->type = BWA_TYPE_NO_MATCH; p->c1 = p->c2 = 0; p->n_aln = 0; p->aln = 0;
-		seq[0] = p->seq; seq[1] = p->rseq;
 		if (max_l < p->len) {
 			max_l = p->len;
-			w[0] = (bwt_width_t*)realloc(w[0], (max_l + 1) * sizeof(bwt_width_t));
-			w[1] = (bwt_width_t*)realloc(w[1], (max_l + 1) * sizeof(bwt_width_t));
-			memset(w[0], 0, (max_l + 1) * sizeof(bwt_width_t));
-			memset(w[1], 0, (max_l + 1) * sizeof(bwt_width_t));
+			w = (bwt_width_t*)realloc(w, (max_l + 1) * sizeof(bwt_width_t));
+			memset(w, 0, (max_l + 1) * sizeof(bwt_width_t));
 		}
-		bwt_cal_width(bwt[0], p->len, seq[0], w[0]);
-		bwt_cal_width(bwt[1], p->len, seq[1], w[1]);
+		bwt_cal_width(bwt, p->len, p->seq, w);
 		if (opt->fnr > 0.0) local_opt.max_diff = bwa_cal_maxdiff(p->len, BWA_AVG_ERR, opt->fnr);
 		local_opt.seed_len = opt->seed_len < p->len? opt->seed_len : 0x7fffffff;
-		if (p->len > opt->seed_len) {
-			bwt_cal_width(bwt[0], opt->seed_len, seq[0] + (p->len - opt->seed_len), seed_w[0]);
-			bwt_cal_width(bwt[1], opt->seed_len, seq[1] + (p->len - opt->seed_len), seed_w[1]);
-		}
+		if (p->len > opt->seed_len)
+			bwt_cal_width(bwt, opt->seed_len, p->seq + (p->len - opt->seed_len), seed_w);
 		// core function
-		p->aln = bwt_match_gap(bwt, p->len, seq, w, p->len <= opt->seed_len? 0 : seed_w, &local_opt, &p->n_aln, stack);
-		// store the alignment
+		for (j = 0; j < p->len; ++j) // we need to complement
+			p->seq[j] = p->seq[j] > 3? 4 : 3 - p->seq[j];
+		p->aln = bwt_match_gap(bwt, p->len, p->seq, w, p->len <= opt->seed_len? 0 : seed_w, &local_opt, &p->n_aln, stack);
+		// clean up the unused data in the record
 		free(p->name); free(p->seq); free(p->rseq); free(p->qual);
 		p->name = 0; p->seq = p->rseq = p->qual = 0;
 	}
-	free(seed_w[0]); free(seed_w[1]);
-	free(w[0]); free(w[1]);
+	free(seed_w); free(w);
 	gap_destroy_stack(stack);
 }
 
 #ifdef HAVE_PTHREAD
 typedef struct {
 	int tid;
-	bwt_t *bwt[2];
+	bwt_t *bwt;
 	int n_seqs;
 	bwa_seq_t *seqs;
 	const gap_opt_t *opt;
@@ -176,20 +156,19 @@ void bwa_aln_core(const char *prefix, const char *fn_fa, const gap_opt_t *opt)
 	bwa_seq_t *seqs;
 	bwa_seqio_t *ks;
 	clock_t t;
-	bwt_t *bwt[2];
+	bwt_t *bwt;
 
 	// initialization
 	ks = bwa_open_reads(opt->mode, fn_fa);
 
 	{ // load BWT
 		char *str = (char*)calloc(strlen(prefix) + 10, 1);
-		strcpy(str, prefix); strcat(str, ".bwt");  bwt[0] = bwt_restore_bwt(str);
-		strcpy(str, prefix); strcat(str, ".rbwt"); bwt[1] = bwt_restore_bwt(str);
+		strcpy(str, prefix); strcat(str, ".bwt");  bwt = bwt_restore_bwt(str);
 		free(str);
 	}
 
 	// core loop
-	fwrite(opt, sizeof(gap_opt_t), 1, stdout);
+	err_fwrite(opt, sizeof(gap_opt_t), 1, stdout);
 	while ((seqs = bwa_read_seq(ks, 0x40000, &n_seqs, opt->mode, opt->trim_qual)) != 0) {
 		tot_seqs += n_seqs;
 		t = clock();
@@ -209,7 +188,7 @@ void bwa_aln_core(const char *prefix, const char *fn_fa, const gap_opt_t *opt)
 			data = (thread_aux_t*)calloc(opt->n_threads, sizeof(thread_aux_t));
 			tid = (pthread_t*)calloc(opt->n_threads, sizeof(pthread_t));
 			for (j = 0; j < opt->n_threads; ++j) {
-				data[j].tid = j; data[j].bwt[0] = bwt[0]; data[j].bwt[1] = bwt[1];
+				data[j].tid = j; data[j].bwt = bwt;
 				data[j].n_seqs = n_seqs; data[j].seqs = seqs; data[j].opt = opt;
 				pthread_create(&tid[j], &attr, worker, data + j);
 			}
@@ -226,8 +205,8 @@ void bwa_aln_core(const char *prefix, const char *fn_fa, const gap_opt_t *opt)
 		fprintf(stderr, "[bwa_aln_core] write to the disk... ");
 		for (i = 0; i < n_seqs; ++i) {
 			bwa_seq_t *p = seqs + i;
-			fwrite(&p->n_aln, 4, 1, stdout);
-			if (p->n_aln) fwrite(p->aln, sizeof(bwt_aln1_t), p->n_aln, stdout);
+			err_fwrite(&p->n_aln, 4, 1, stdout);
+			if (p->n_aln) err_fwrite(p->aln, sizeof(bwt_aln1_t), p->n_aln, stdout);
 		}
 		fprintf(stderr, "%.2f sec\n", (float)(clock() - t) / CLOCKS_PER_SEC); t = clock();
 
@@ -236,7 +215,7 @@ void bwa_aln_core(const char *prefix, const char *fn_fa, const gap_opt_t *opt)
 	}
 
 	// destroy
-	bwt_destroy(bwt[0]); bwt_destroy(bwt[1]);
+	bwt_destroy(bwt);
 	bwa_seq_close(ks);
 }
 
@@ -246,7 +225,7 @@ int bwa_aln(int argc, char *argv[])
 	gap_opt_t *opt;
 
 	opt = gap_init_opt();
-	while ((c = getopt(argc, argv, "n:o:e:i:d:l:k:cLR:m:t:NM:O:E:q:f:b012IB:")) >= 0) {
+	while ((c = getopt(argc, argv, "n:o:e:i:d:l:k:cLR:m:t:NM:O:E:q:f:b012IYB:")) >= 0) {
 		switch (c) {
 		case 'n':
 			if (strstr(optarg, ".")) opt->fnr = atof(optarg), opt->max_diff = -1;
@@ -274,6 +253,7 @@ int bwa_aln(int argc, char *argv[])
 		case '1': opt->mode |= BWA_MODE_BAM_READ1; break;
 		case '2': opt->mode |= BWA_MODE_BAM_READ2; break;
 		case 'I': opt->mode |= BWA_MODE_IL13; break;
+		case 'Y': opt->mode |= BWA_MODE_CFY; break;
 		case 'B': opt->mode |= atoi(optarg) << 24; break;
 		default: return 1;
 		}
@@ -303,7 +283,7 @@ int bwa_aln(int argc, char *argv[])
 		fprintf(stderr, "         -q INT    quality threshold for read trimming down to %dbp [%d]\n", BWA_MIN_RDLEN, opt->trim_qual);
         fprintf(stderr, "         -f FILE   file to write output to instead of stdout\n");
 		fprintf(stderr, "         -B INT    length of barcode\n");
-		fprintf(stderr, "         -c        input sequences are in the color space\n");
+//		fprintf(stderr, "         -c        input sequences are in the color space\n");
 		fprintf(stderr, "         -L        log-scaled gap penalty for long deletions\n");
 		fprintf(stderr, "         -N        non-iterative mode: search for all n-difference hits (slooow)\n");
 		fprintf(stderr, "         -I        the input is in the Illumina 1.3+ FASTQ-like format\n");
@@ -311,6 +291,7 @@ int bwa_aln(int argc, char *argv[])
 		fprintf(stderr, "         -0        use single-end reads only (effective with -b)\n");
 		fprintf(stderr, "         -1        use the 1st read in a pair (effective with -b)\n");
 		fprintf(stderr, "         -2        use the 2nd read in a pair (effective with -b)\n");
+		fprintf(stderr, "         -Y        filter Casava-filtered sequences\n");
 		fprintf(stderr, "\n");
 		return 1;
 	}
