@@ -22,6 +22,8 @@
 #define BWA_AVG_ERR 0.02
 #define BWA_MIN_RDLEN 35 // for read trimming
 
+#define BWA_MAX_BCLEN 63 // maximum barcode length; 127 is the maximum
+
 #ifndef bns_pac
 #define bns_pac(pac, k) ((pac)[(k)>>2] >> ((~(k)&3)<<1) & 3)
 #endif
@@ -32,7 +34,7 @@ typedef struct {
 } bwt_width_t;
 
 typedef struct {
-	uint32_t n_mm:8, n_gapo:8, n_gape:8, a:1;
+	uint32_t n_mm:16, n_gapo:8, n_gape:8;
 	bwtint_t k, l;
 	int score;
 } bwt_aln1_t;
@@ -49,8 +51,8 @@ typedef uint16_t bwa_cigar_t;
 #define __cigar_create(__op, __len) ((__op)<<CIGAR_OP_SHIFT | (__len))
 
 typedef struct {
-	uint32_t pos;
 	uint32_t n_cigar:15, gap:8, mm:8, strand:1;
+	bwtint_t pos;
 	bwa_cigar_t *cigar;
 } bwt_multi1_t;
 
@@ -75,7 +77,7 @@ typedef struct {
 	// for multi-threading only
 	int tid;
 	// barcode
-	char bc[16]; // null terminated; up to 15 bases
+	char bc[BWA_MAX_BCLEN+1]; // null terminated; up to BWA_MAX_BCLEN bases
 	// NM and MD tags
 	uint32_t full_len:20, nm:12;
 	char *md;
@@ -84,6 +86,7 @@ typedef struct {
 #define BWA_MODE_GAPE       0x01
 #define BWA_MODE_COMPREAD   0x02
 #define BWA_MODE_LOGGAP     0x04
+#define BWA_MODE_CFY        0x08
 #define BWA_MODE_NONSTOP    0x10
 #define BWA_MODE_BAM        0x20
 #define BWA_MODE_BAM_SE     0x40
@@ -132,7 +135,7 @@ extern "C" {
 	void bwa_free_read_seq(int n_seqs, bwa_seq_t *seqs);
 
 	int bwa_cal_maxdiff(int l, double err, double thres);
-	void bwa_cal_sa_reg_gap(int tid, bwt_t *const bwt[2], int n_seqs, bwa_seq_t *seqs, const gap_opt_t *opt);
+	void bwa_cal_sa_reg_gap(int tid, bwt_t *const bwt, int n_seqs, bwa_seq_t *seqs, const gap_opt_t *opt);
 
 	void bwa_cs2nt_core(bwa_seq_t *p, bwtint_t l_pac, ubyte_t *pac);
 
